//+------------------------------------------------------------------+
//|                                                       Copier.mq5 |
//|                              Copyright  2016, Vladimir Karputov |
//|                                           http://wmua.ru/slesar/ |
//+------------------------------------------------------------------+
#property copyright "Copyright  2016, Vladimir Karputov"
#property link      "http://wmua.ru/slesar/"
#property version   "2.110"
#property description "The expert for copying of signals" 
#include <Trade\Trade.mqh>
#include "Languages.mqh"
//--- input parameters 
input int  ExtMagicNumber=2047483647;  // Expert's identifier 
input uchar ExtCopyRatio=5;            // Deals multiply by
//---
bool  hedging=true;
string my_comment="";
//+------------------------------------------------------------------+
//| Structure of congruences of positions of the terminal and        |
//| positions of the copier                                          |
//+------------------------------------------------------------------+
struct correlation
  {
   long              POSITION_IDENTIFIER_terminal; // Position identifier (terminal)
   //+------------------------------------------------------------------+
   //| Position identifier is a unique number that is assigned          |
   //| to every newly opened position and doesn't change during         |
   //| the entire lifetime of the position.                             |
   //| Position turnover doesn't change its identifier.                 |
   //+------------------------------------------------------------------+
   double            POSITION_VOLUME_terminal;     // Volume of the position opened by the Signals service
   long              POSITION_IDENTIFIER_copier;   // Position identifier (copier)
   //+------------------------------------------------------------------+
   //| Position identifier is a unique number that is assigned          |
   //| to every newly opened position and doesn't change during         |
   //| the entire lifetime of the position.                             |
   //| Position turnover doesn't change its identifier.                 |
   //+------------------------------------------------------------------+
   ulong             DEAL_ticket;                  // Deal ticket, if the deal is executed.
  };
correlation       arr_correlation[];
CLng              m_languages;                     // the languages object
//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
int OnInit()
  {
//--- create timer
   EventSetTimer(600);
//---
   my_comment=MQLInfoString(MQL_PROGRAM_NAME);
//---
   ENUM_ACCOUNT_MARGIN_MODE m_margin_mode;                  // netting or hedging
   m_margin_mode=(ENUM_ACCOUNT_MARGIN_MODE)AccountInfoInteger(ACCOUNT_MARGIN_MODE);
   if(m_margin_mode==ACCOUNT_MARGIN_MODE_RETAIL_HEDGING)
     {
      hedging=true;
      //Print("hedging selected");
     }
   else if(m_margin_mode==ACCOUNT_MARGIN_MODE_RETAIL_NETTING)
      hedging=false;
//---
   ArrayResize(arr_correlation,0);
//---
   Print(m_languages.GetText(0),(bool)SignalInfoGetInteger(SIGNAL_INFO_CONFIRMATIONS_DISABLED));
   Print(m_languages.GetText(1),(bool)SignalInfoGetInteger(SIGNAL_INFO_COPY_SLTP));
   Print(m_languages.GetText(2),SignalInfoGetInteger(SIGNAL_INFO_DEPOSIT_PERCENT));
   Print("id ",m_languages.GetText(3),"https://www.mql5.com/",
         m_languages.GetText(4),"/signals/",SignalInfoGetInteger(SIGNAL_INFO_ID));
   Print(m_languages.GetText(5),(bool)SignalInfoGetInteger(SIGNAL_INFO_SUBSCRIPTION_ENABLED));
   Print(m_languages.GetText(6),(bool)SignalInfoGetInteger(SIGNAL_INFO_TERMS_AGREE));
//---
   return(INIT_SUCCEEDED);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
//--- destroy timer
   EventKillTimer();
  }
//+------------------------------------------------------------------+
//| Expert tick function                                             |
//+------------------------------------------------------------------+
void OnTick()
  {
//---

  }
//+------------------------------------------------------------------+
//| Timer function                                                   |
//+------------------------------------------------------------------+
void OnTimer()
  {
//---

  }
//+------------------------------------------------------------------+ 
//| TradeTransaction function                                        | 
//+------------------------------------------------------------------+ 
void OnTradeTransaction(const MqlTradeTransaction &trans,
                        const MqlTradeRequest &request,
                        const MqlTradeResult &result)
  {
//--- get transaction type as enumeration value 
   ENUM_TRADE_TRANSACTION_TYPE type=trans.type;
//--- if transaction is result of addition of the transaction in history
   if(type==TRADE_TRANSACTION_DEAL_ADD)
     {
      long     deal_entry        =0;
      double   deal_volume       =0;
      string   deal_symbol       ="";
      long     deal_type         =0;
      long     deal_magic        =0;
      long     deal_positions_id =0;
      string   deal_comment      ="";
      if(HistoryDealSelect(trans.deal))
        {
         deal_entry=HistoryDealGetInteger(trans.deal,DEAL_ENTRY);
         deal_volume=HistoryDealGetDouble(trans.deal,DEAL_VOLUME);
         deal_symbol=HistoryDealGetString(trans.deal,DEAL_SYMBOL);
         deal_type=HistoryDealGetInteger(trans.deal,DEAL_TYPE);
         deal_magic=HistoryDealGetInteger(trans.deal,DEAL_MAGIC);
         deal_positions_id=HistoryDealGetInteger(trans.deal,DEAL_POSITION_ID);
         deal_comment=HistoryDealGetString(trans.deal,DEAL_COMMENT);
        }
      else
         return;
      if(deal_entry==DEAL_ENTRY_IN)
        {
         int index_row=-1;
         if(!FindDealTicket(index_row,trans.deal))
           {
            int size=ArraySize(arr_correlation);
            ArrayResize(arr_correlation,size+1);
            size=ArraySize(arr_correlation);
            FillingCorrelation(size-1,0,0,0,0,true);
            PrintBase();
            //---
            ulong rezult=0;
            if(OpenDeal(deal_volume,deal_symbol,deal_type,deal_magic,rezult))
              {
               FillingCorrelation(size-1,deal_positions_id,deal_volume,0,rezult,false);
               PrintBase();
              }
           }
         else
           {
            //---   -     , 
            //---         DEAL_POSITOIN_ID. 
            //---   DEAL_POSITOIN_ID   POSITION_IDENTIFIER copier.
            //--- if we find - it means the transaction opened by the copier,
            //--- therefore search of the transaction in the history and receiving her DEAL_POSITOIN_ID.
            //--- we write down this DEAL_POSITOIN_ID in the POSITION_IDENTIFIER copier element.
            //Print(" DEAL_POSITOIN_ID   POSITION_IDENTIFIER copier");
            FillingCorrelation(index_row,0,0,deal_positions_id,0,false);
            PrintBase();
           }
        }
      else if(deal_entry==DEAL_ENTRY_OUT)
        {
         if(hedging)
           {
            //Print("hedging \"OUT\"");
            int   index_row         =-1;
            bool  find_id_terminal  =false;
            bool  find_id_copier    =false;
            if(FindDEAL_POSITION_IdInStructure(index_row,find_id_terminal,find_id_copier,deal_positions_id))
              {
               //Print(deal_positions_id,"    ",index_row,
               //      "  POSITION_IDENTIFIER_terminal ",find_id_terminal,
               //      " POSITION_IDENTIFIER_copier ",find_id_copier);
               long POSITION_IDENTIFIER_terminal=arr_correlation[index_row].POSITION_IDENTIFIER_terminal;
               long POSITION_IDENTIFIER_copier=arr_correlation[index_row].POSITION_IDENTIFIER_copier;
               PrintBase();
               if(find_id_copier)
                 {
                  return;
                 }
               if(find_id_terminal)
                 {
                  double   vol_terminal   =0;
                  long     ticket_terminal=0;
                  string   symbol_terminal="";
                  if(FindPosition(vol_terminal,ticket_terminal,symbol_terminal,POSITION_IDENTIFIER_terminal))
                    {
                     //Print("  ",POSITION_IDENTIFIER_terminal," ,   ",vol_terminal,
                     //      ",   ",ticket_terminal);
                     CTrade copier_trade;
                     if(!hedging)
                        copier_trade.SetExpertMagicNumber(0);
                     else
                        copier_trade.SetExpertMagicNumber(ExtMagicNumber);
                     copier_trade.PositionClose(POSITION_IDENTIFIER_copier);
                     copier_trade.ResultDeal();
                     //Print("   . ResultDeal()=",copier_trade.ResultDeal());
                     if(copier_trade.ResultDeal()!=0)
                       {
                        ulong rezult=0;
                        double volume=0;
                        if(OpenDeal(vol_terminal,deal_symbol,deal_type,deal_magic,rezult))
                          {
                           FillingCorrelation(index_row,0,0,0,rezult,false);
                           PrintBase();
                          }
                       }
                     return;
                    }
                  else
                    {
                     //Print("  ",POSITION_IDENTIFIER_terminal,"  ");
                     CTrade copier_trade;
                     if(!hedging)
                        copier_trade.SetExpertMagicNumber(0);
                     else
                        copier_trade.SetExpertMagicNumber(ExtMagicNumber);
                     copier_trade.PositionClose(POSITION_IDENTIFIER_copier);
                     copier_trade.ResultDeal();
                     //Print("   . ResultDeal()=",copier_trade.ResultDeal());
                     return;
                    }
                 }
              }
            else
              {
               //Print(deal_positions_id,"  ");
               return;
              }
           }
         if(!hedging)
           {
            if(deal_comment==my_comment)
               return;
            //Print("!hedging \"OUT\"");
            int index_row=-1;
            if(FindDEAL_POSITION_IdInStructure(index_row,deal_positions_id))
              {
               //Print(deal_positions_id,"    ",index_row);
               long POSITION_IDENTIFIER_terminal=arr_correlation[index_row].POSITION_IDENTIFIER_terminal;
               PrintBase();
               double   vol_terminal=0;
               long     ticket_terminal=0;
               string   symbol_terminal="";
               if(FindPosition(vol_terminal,ticket_terminal,symbol_terminal,POSITION_IDENTIFIER_terminal))
                 {
                  //Print("  ",POSITION_IDENTIFIER_terminal," ,   ",vol_terminal,
                  //      ",   ",ticket_terminal);
                  double   new_volume=arr_correlation[index_row].POSITION_VOLUME_terminal-deal_volume;
                  //Print("new_volume=",DoubleToString(new_volume,2));
                  long     new_type    =0;
                  ulong    new_ticket  =0;
                  if((ENUM_DEAL_TYPE)deal_type==DEAL_TYPE_BUY)
                     new_type=DEAL_TYPE_SELL;
                  if((ENUM_DEAL_TYPE)deal_type==DEAL_TYPE_SELL)
                     new_type=DEAL_TYPE_BUY;
                  if(new_volume<=0)
                    {
                     if(OpenDeal(vol_terminal,deal_symbol,new_type,deal_magic,new_ticket))
                       {
                        FillingCorrelation(index_row,0,0,0,new_ticket,false);
                        PrintBase();
                       }
                    }
                  else
                     OpenDeal(deal_volume,deal_symbol,deal_type,deal_magic,new_ticket);
                 }
              }
            else
              {
               //Print(deal_positions_id,"  ");
               return;
              }
           }
        }
     }
  }
//+------------------------------------------------------------------+
//| Filling of elements of structure                                 |
//+------------------------------------------------------------------+
void FillingCorrelation(const int index,long position_identifier_terminal=0,
                        const double position_volume_terminal=0,
                        long position_identifier_copier=0,ulong deal_ticket=0,
                        bool init=false)
  {
   if(init)
     {
      arr_correlation[index].POSITION_IDENTIFIER_terminal=0;
      arr_correlation[index].POSITION_VOLUME_terminal=0;
      arr_correlation[index].POSITION_IDENTIFIER_copier=0;
      arr_correlation[index].DEAL_ticket=0;
      return;
     }
   if(position_identifier_terminal!=0)
      arr_correlation[index].POSITION_IDENTIFIER_terminal=position_identifier_terminal;
   if(position_volume_terminal!=0)
      arr_correlation[index].POSITION_VOLUME_terminal=position_volume_terminal;
   if(position_identifier_copier!=0)
      arr_correlation[index].POSITION_IDENTIFIER_copier=position_identifier_copier;
   if(deal_ticket!=0)
      arr_correlation[index].DEAL_ticket=deal_ticket;
  }
//+------------------------------------------------------------------+
//| Print all correlation                                            |
//+------------------------------------------------------------------+
void PrintBase(void)
  {
   return;
   int size=ArraySize(arr_correlation);
   for(int i=size-1;i>=0;i--)
     {
      Print(i,": ",
            arr_correlation[i].POSITION_IDENTIFIER_terminal,", ",
            DoubleToString(arr_correlation[i].POSITION_VOLUME_terminal,2),", ",
            arr_correlation[i].POSITION_IDENTIFIER_copier,", ",
            arr_correlation[i].DEAL_ticket);
     }
  }
//+------------------------------------------------------------------+
//| Search in the DEAL_ticket  element                               |
//+------------------------------------------------------------------+
bool FindDealTicket(int &index,const ulong deal_ticket)
  {
   for(int i=ArraySize(arr_correlation)-1;i>=0;i--)
     {
      if(arr_correlation[i].DEAL_ticket==deal_ticket)
        {
         index=i;
         return(true);
        }
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| Opening of the deal of the increased size                        |
//+------------------------------------------------------------------+
bool OpenDeal(const double volume,const string symbol,const long type,const long magic,ulong &ticket)
  {
   double vol=volume*ExtCopyRatio;
   //Print("  ",EnumToString((ENUM_DEAL_TYPE)type));
//---
   CTrade copier_trade;
   if(!hedging)
      copier_trade.SetExpertMagicNumber(magic);
   else
      copier_trade.SetExpertMagicNumber(ExtMagicNumber);
   if((ENUM_DEAL_TYPE)type==DEAL_TYPE_BUY)
     {
      copier_trade.Buy(vol,symbol,0,0,0,my_comment);
      if(copier_trade.ResultDeal()!=0)
        {
         ticket=copier_trade.ResultDeal();
         //Print(" Buy ",symbol," ",DoubleToString(vol,2),", ticket ",ticket);
         return(true);
        }
     }
   if((ENUM_DEAL_TYPE)type==DEAL_TYPE_SELL)
     {
      copier_trade.Sell(vol,symbol,0,0,0,"Copier");
      if(copier_trade.ResultDeal()!=0)
        {
         ticket=copier_trade.ResultDeal();
         //Print(" Sell ",symbol," ",DoubleToString(vol,2),", ticket ",ticket);
         return(true);
        }
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| Find DEAL_POSITION_ID in elements POSITION_IDENTIFIER_terminal   |
//| and POSITION_IDENTIFIER_copier.                                  |
//+------------------------------------------------------------------+
bool FindDEAL_POSITION_IdInStructure(int &index,const long deal_POSITION_ID=0)
  {
//---        -   ID
//---      POSITION_IDENTIFIER_terminal
//--- at a netting it is senseless to look for on both elements - there identical ID
//--- therefore we will look for on the POSITION_IDENTIFIER_terminal element
   if(!hedging)
     {
      for(int i=ArraySize(arr_correlation)-1;i>=0;i--)
        {
         if(arr_correlation[i].POSITION_IDENTIFIER_terminal==deal_POSITION_ID)
           {
            index=i;
            return(true);
           }
        }
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| Find DEAL_POSITION_ID in elements POSITION_IDENTIFIER_terminal   |
//| and POSITION_IDENTIFIER_copier.                                  |
//+------------------------------------------------------------------+
bool FindDEAL_POSITION_IdInStructure(int &index,bool &terminal,bool &copier,const long deal_POSITION_ID=0)
  {
//---       
//--- at a hedzhing we will look for on both elements
   if(hedging)
     {
      for(int i=ArraySize(arr_correlation)-1;i>=0;i--)
        {
         if(arr_correlation[i].POSITION_IDENTIFIER_terminal==deal_POSITION_ID)
           {
            index=i;
            terminal=true;
            return(true);
           }
         if(arr_correlation[i].POSITION_IDENTIFIER_copier==deal_POSITION_ID)
           {
            index=i;
            copier=true;
            return(true);
           }
        }
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| Search of a position on her POSITION_IDENTIFIER                  |
//+------------------------------------------------------------------+
bool FindPosition(double &volume,long &ticket,string &symbol,const long position_identifier)
  {
   for(int i=0;i<PositionsTotal();i++)
     {
      symbol=PositionGetSymbol(i);
      if(PositionGetInteger(POSITION_IDENTIFIER)==position_identifier)
        {
         volume=PositionGetDouble(POSITION_VOLUME);
         ticket=PositionGetInteger(POSITION_TICKET);
         return(true);
        }
     }
   return(false);
  }
//+------------------------------------------------------------------+
